;*****************************************************************************************
;
;       PROGRAM:                DD_SYNTH.ASM (SOURCE FILE)
;                               DD_SYNTH.DAT (EEPROM FREQUENCY SETTINGS ETC.)
;
;       ASSEMBLER:              MPASMWIN v02.50.02
;
;       AUTHOR:                 Steven Jones         EMAIL stevejones@picknowl.com.au
;
;
;       REVISION HISTORY:
;
;       VERSION:        DATE            COMMENTS
;       1.00            NOV 09 2000     FIRST RELEASE.
;       1.10            JAN 20 2001     MOD TO OFFSET FREQUENCY SETUP SCREEN
;                                       TO ALLOW NEGATIVE OFFSETS.
;
;
;       GENERAL COMMENTS
;
;       This program is based on the various versions of SIG_GEN,
;       by Curtis W. Preuss, WB2V. And modified and updated by among others :-
;       Bruce Stough,    - AA0ED
;       Craig Johnson,   - AA0ZZ
;       George Heron,    - N2APB
;
;       The transceiver functions, TX offset & Receive Incremental Tuning (RIT)
;       are based on modifications by Mark Kilmier -VK5EME to SIG_GEN version 3a, which
;       Mark has named SIGGEN2.
;
;       The aim of this project was to include as many of the following functions
;       suggested by Mark, as possible.
;
;       1) Use the original PCB, modified to gain access to RA2..RA4 pins. Using 
;          diodes to provide the number of switch inputs required.
;       2) DDS frequency output, 0Hz to 40MHz.
;       3) TX/RX switching. By grounding RA4 to go to TX mode.
;       4) Frequency readout can be offset by 1Hz to 11GHz.
;       5) No requirement for band switching, but able to program upper and lower limits.
;       6) Memorise the last frequency on power up.
;       7) Use low cost mechanical rotary encoder instead of the optical type.
;       8) Variable rate tuning not required.
;       9) By grounding RA2, (PIC pin 1) then turning the rotary encoder, be able to
;          set the step size required. eg 100Hz, 1KHz, 10KHz etc.
;      10) Use 16x2 LCD with or without R/W pin. (use delays rather than busy checks)
;      11) Software fine tuning of the XTAL oscillator frequency still required.
;      12) RIT available in RX mode by grounding, via diodes, both RA3 & RA4.
;      13) Still require the option of programming the software for different XTAL
;          oscillators. eg 66MHz, 100MHz, 125MHz.
;
;       
;       My implementation.
;
;       1) Uses the include file DD_SYNTH.DAT to set the following EEPROM defaults.
; 
;          DDS_XTAL_FREQ.   range 30 MHz to 130 MHz. (eg, Set to 66000000 for 66MHz) 
;          MAX_DDS_FREQ.    range  0 MHz to  60 MHz. (Usually 1/3 DDS XTAL FREQ)
;          RX_DDS_FREQ.
;          MIN_RX_DDS_FREQ.
;          MAX_RX_DDS_FREQ.
;          MAX_RIT_OFFSET.
;          TX_OFFSET_FREQ.              (Offset from RX frequency)
;          OFFSET_FREQ.     20 GHz max. (Added to TX/RX frequency before display on LCD)
;          MULTIPLIER.         1 to 20. (Frequency display is multiplied by N. Limit N )
;                                       (so result is not to much greater than 20 GHz. )
;
;          (note. no checks are made to make sure the settings are valid)
;
;          The frequency displayed on the LCD, is calculated as shown below.
;          RX = MULTIPLIER x (RX_DDS_FREQ + OFFSET_FREQ)
;          TX = MULTIPLIER x (RX_DDS_FREQ + TX_OFFSET_FREQ + OFFSET_FREQ)
;
;          The frequency programmed into the DDS, is calculated as shown below.
;          RX = ABS(RX_DDS_FREQ)
;          TX = ABS(RX_DDS_FREQ + TX_OFFSET_FREQ)
;
;          RX_DDS_FREQ, MIN_RX_DDS_FREQ, MAX_RX_DDS_FREQ, TX_OFFSET_FREQ & OFFSET_FREQ
;          may be positive or negative values.
;          (as long as the resulting display freq is positive)
;
;       2) Once the PIC is programmed, all of the above defaults can be modified by
;          entering the setup screens. Changes can be made using the rotary encoder
;          and saved to EEPROM. All setup screens check the changes to make sure they
;          are valid. eg The TX offset cannot be set to a value that when added to the
;          RX frequency, results in a frequency outside the range of the DDS.
;
;       3) When first turned on, the software name and version number is displayed.
;
;       4) After 1 second the main screen is displayed. 
;          The frequency displayed is the same as the last frequency displayed on the
;          main screen before the power was turned off.
;          If the RA4 (TX) pin is pulled low TX is displayed, otherwise RX is displayed.
;          Any changes to the RX frequency are saved to EEPROM 2 seconds after the rotary
;          encoder stops moving. 
;          If RA3 & RA4 via diodes (the RIT button) is briefly pulled low, RIT is 
;          displayed on the LCD. The TX frequency cannot be changed. But the RIT frequency
;          can be changed as long as it does not go to far away from the RX frequency. 
;          (the range is RX freq +/-MAX_RIT_OFFSET) Pressing the RIT button again, removes
;          the RIT display and the RX frequency reverts to its pre RIT value.
;
;       5) To enter the setup screens, pull RA3 (CAL button) low, while turning on the
;          power. After the version number is displayed the first setup screen, for
;          setting the DDS XTAL freq, is displayed. 
;          It sets the DDS to produce 10MHz, and displays the DDS calibration constant. 
;          This can be modified to set the boards O/P frequency to exactly 10 MHz.
;          (using a frequency counter)
;          The DDS calibration constant is equal to 2^56/(DDS XTAL freq). So if you change
;          to a different XTAL frequency you can change the constant as shown below.
;
;          For  50.00 MHz,  2^56/50.00 MHz = 1441151880
;               66.00 MHz,  2^56/66.00 MHz = 1091781727
;               66.66 MHz,  2^56/66.66 MHz = 1080972007
;              100.00 MHz,  2^56/100.0 MHz =  720575940
;              120.00 MHz,  2^56/120.0 MHz =  600479950
;
;          This screen will continue to be displayed until the CAL button is pressed
;          briefly. This will save the changes to EEPROM and display the next settings
;          screen.
;          Again, pressing the CAL button will save any changes made on this screen to
;          EEPROM and display the next settings screen. etc.
;          On all setup screens except the first one, if the buttons and rotary encoder
;          are left idle for 10 seconds, any changes made on that screen will be ignored
;          and you will be returned to the main display.
;          Once in the main screen (not RIT) pressing the CAL button for 1 sec will
;          take you to the 3rd setup screen. (the first two screens rarely need changing)
;
;       6) If RA2 (adjust step size button) is pulled low, a cursor is displayed under one
;          digit of the frequency display. This can be changed using the rotary encoder.
;          If for example it is under the 1 KHz digit, when the (adjust step size button)
;          is released, rotating the encoder will change the frequency in 1 KHz steps.
;          There is also a small amount of variable rate tuning. (the faster the encoder
;          is rotated, the greater the step size) Due to 4mS software debouncing, if the
;          encoder is rotated to fast, no change in frequency will occur.
;          Any changes to the step size, while in the main screen, are saved to EEPROM.
;
;       7) Uses a 16x2 LCD with or without R/W pin. (use delays rather than busy checks)
;
;       8) To enable all of the required functions to be programmed into the 1K of code 
;          space, any code that was used more than once or twice was changed to a
;          subroutine. And wherever a subroutine ended with a CALL and RETURN, eg
;               NOP
;               CALL    XXXXXX
;               RETURN
;
;          The code was changed to a GOTO.
;               NOP
;               GOTO    XXXXXX                  ; Return via "GOTO".
;
;          I know this is poor programming practice, but it did save a lot of code space.
;          Each of these modifications as identified by the [Return via "GOTO".] comment.
;
;*****************************************************************************************

;*****************************************************************************************
;               Register definitions
;*****************************************************************************************

W               EQU             0
F               EQU             1

INDF            SET             H'00'
TMR0            SET             H'01'
PCL             SET             H'02'
STATUS          SET             H'03'
FSR             SET             H'04'
PORTA           SET             H'05'
PORTB           SET             H'06'
EEDATA          SET             H'08'
EEADR           SET             H'09'
INTCON          SET             H'0B'
OPTION_REG      SET             H'01'
EECON1          SET             H'08'
EECON2          SET             H'09'
#DEFINE         CARRY           STATUS,0
#DEFINE         ZERO            STATUS,2
#DEFINE         RP0             STATUS,5
#DEFINE         T0IF            INTCON,2
#DEFINE         RD              EECON1,0
#DEFINE         WR              EECON1,1
#DEFINE         WREN            EECON1,2
#DEFINE         WRERR           EECON1,3
#DEFINE         EEIF            EECON1,4

;*****************************************************************************************
;               Configuration bits
;*****************************************************************************************

_CP_ON          EQU     H'3FEF'
_CP_OFF         EQU     H'3FFF'
_PWRTE_ON       EQU     H'3FF7'
_PWRTE_OFF      EQU     H'3FFF'
_WDT_OFF        EQU     H'3FFB'
_XT_OSC         EQU     H'3FFD'

        PROCESSOR       16F84                   ; Select processor, and assembler options.
        LIST            F=INHX8M,N=0
        RADIX           DEC

        __CONFIG        _XT_OSC & _WDT_OFF      ; XTAL oscillator, no watchdog.
        __MAXRAM H'CF'
        __BADRAM H'07', H'50'-H'7F', H'87'

;*****************************************************************************************
;               General definitions
;*****************************************************************************************

#DEFINE         ENCODER_A       PORTA,0         ; Rotary encoder switch.
#DEFINE         ENCODER_B       PORTA,1         ; Rotary encoder switch.
#DEFINE         SWITCH_A        PORTA,2         ; External switch array . (active low)
#DEFINE         SWITCH_B        PORTA,3         ;                 "
#DEFINE         SWITCH_C        PORTA,4         ;                 "

#DEFINE         LCD_RS          PORTB,1         ; LCD Reg select. 0 = instruct 1 = data
#DEFINE         LCD_RW          PORTB,2         ; LCD Read/Write  0 = WR 1 = RD (not used)
#DEFINE         LCD_E           PORTB,3         ; LCD Enable.     0 = disable  1 = enable
;                               PORTB,4..7      ; LCD 4 bit data bus.
#DEFINE         DDS_LOAD        PORTB,0         ; AD9850 update pin.
#DEFINE         DDS_CLOCK       PORTB,5         ; AD9850 write clock pin.
#DEFINE         DDS_DATA        PORTB,7         ; AD9850 serial data input pin.

#DEFINE         ENC_A_DEBOUNCED FLAGS,0         ; Encoder A state after debouncing.
#DEFINE         ENC_B_DEBOUNCED FLAGS,1         ; Encoder B state after debouncing.
#DEFINE         OLD_A_DEBOUNCED FLAGS,2         ; Old encoder a state after debouncing.
#DEFINE         LEADING_FLAG    FLAGS,3         ; Leading zero indicator.
#DEFINE         NEGATIVE        FLAGS,4         ; Number to display is negative.
#DEFINE         NEGATIVE_RX     FLAGS,5         ; Min RX DDS Freq is negative.
#DEFINE         TIMEOUT         FLAGS,6         ; If set, go back to main display.
#DEFINE         EEPROM_UPDATE   FLAGS,7         ; If set, update the EEPROM.

#DEFINE         CAL_SCRN        FLAG2,0         ; Set if displaying 10 MHz CAL screen.
#DEFINE         RIT_SCRN        FLAG2,1         ; set if displaying RIT screen.
#DEFINE         PRESSED         FLAG2,2         ; Result flag from button pressed call.


;*****************************************************************************************
;               General equates
;*****************************************************************************************

RESETVECTOR     EQU             H'00'           ; PIC16F84 Reset vector.
INTVECTOR       EQU             H'04'           ; PIC16F84 Interrupt vector.
TRUE            EQU             H'FF'
FALSE           EQU             H'00'

MIN_ON          EQU              6              ; Min valid button pressed time. (6 x 8mS)
MAX_ON          EQU             70              ; Max valid button pressed time.(70 x 8mS)
TIMEOUT_NUM     EQU              5              ; Display timeout. 10 sec (5 x 2.1 sec)

;*****************************************************************************************
                ORG     H'2100'

                INCLUDE "DD_SYNTH.DAT"          ; File to set the initial DDS & display
                                                ; frequency range into EEPROM.
;*****************************************************************************************



;*****************************************************************************************
;               RAM variables 68 bytes max.
;*****************************************************************************************

                CBLOCK  H'0C'                   ; Ram starts at H'0C'.

                SAVE_STATUS                     ; For interrupt service routine.
                SAVE_W_REG                      ;             "
                ENC_HISTORY                     ; Last 4 samples of encoder state.
                COUNT_0                         ; Counter for delay routine.
                COUNT_1                         ; General counter variables.

                COUNT_2                         ; ** All variables from here to the end **
                COUNT_3                         ;    of RAM are cleared at power up.
                COUNT_4                         ;
                TEMP_1                          ; Temporary storage.
                TEMP_2                          ;        "
                TEMP_3                          ;        "
                ENC_COUNTER                     ; 1 to 38 = CW. -1 to -38 = CCW.
                INT_COUNT                       ; 8 to 1. Decremented every 1.024 mS.
                SW_BC_CNT                       ; RIT switch pressed counter. (x 8mS)
                SW_BC_OLD                       ; State of old switch count.
                SW_B_CNT                        ; CAL switch pressed counter. (x 8mS)
                SW_B_OLD                        ; State of old switch count.
                SW_C_CNT                        ; TX switch pressed counter.  (x 8mS)
                TIMEOUT_L                       ; 16 bit display timeout down counter,
                TIMEOUT_H                       ; return to main display when = 0.
                EEPROM_WRITE                    ; EEPROM write update down counter.
                FLAGS                           ; 8 misc flags.
                FLAG2                           ;       "

                FREQUENCY:5                     ; 40 bit frequency.
                STEP_NUM                        ; LCD cursor pos & step size num.(0-7)
                                                ; (Do not separate frequency & STEP_SIZE)
                STEP_RATE                       ;
                DDS_OLD:5                       ; 40 bit previous DDS freq.

                ARG1_10                         ; MSD.    11 byte ASCII buffer.
                ARG1_9                          ; MSD. \
                ARG1_8                          ;       |
                ARG1_7                          ;       |
                ARG1_6                          ;       |
                ARG1_5                          ;       | 80 bit maths buffer.
                ARG1_4                          ;       |
                ARG1_3                          ;       |
                ARG1_2                          ;       |
                ARG1_1                          ;       |
                ARG1_0                          ; LSD. /

                ARG2_4                          ; MSD. \
                ARG2_3                          ;       |
                ARG2_2                          ;       | 40 bit maths buffer.
                ARG2_1                          ;       |
                ARG2_0                          ; LSD. /

                ARG3_4                          ; MSD. \
                ARG3_3                          ;       |
                ARG3_2                          ;       | 40 bit maths buffer.
                ARG3_1                          ;       |
                ARG3_0                          ; LSD. /

                ARG4_4                          ; MSD. \
                ARG4_3                          ;       |
                ARG4_2                          ;       | 40 bit maths buffer.
                ARG4_1                          ;       |
                ARG4_0                          ; LSD. /

                ENDC

;*****************************************************************************************
;
;               Initialise the hardware.
;
;*****************************************************************************************

                ORG     RESETVECTOR

                BSF     RP0                     ; Select page 1 for TRIS reg access.
                MOVLW   B'00011111'
                MOVWF   PORTA                   ; RA0..RA4 = inputs.
                GOTO    CONTINUE                ; Jump over interrupt routines.

;*****************************************************************************************
;
;               Interrupt service routine
;
;               Service the TMR0 interrupt (EVERY 1.024 mS)
;
;               EVERY 1.024 mS.
;               Debounce the rotary encoder by sampling the encoder state every mS.
;               If 4 consecutive checks of the pins are at the same state, that state
;               is considered to be valid. (debounced)
;               Determine the direction of rotation by looking at the debounced
;               Encoder B pin on the rising edge of the debounced encoder A pin.
;               Encoder B pin low = Clockwise, high = Counter-Clockwise.
;               On each determination of the direction, set the ENC_COUNTER variable
;               to the current state of the STEP_RATE variable.
;               (+ STEP_RATE for CW, - STEP_RATE for CCW)
;               Then set the STEP_RATE variable to 38.
;               This variable is then decremented toward 1 every mS. Therefore the step
;               rate increases from 1, for slow rotation, up to 30 for fast rotation.
;               (The variable would have decremented to at least 30 before)
;               (the next rising edge of the debounced encoder A pin.     )
;
;               EVERY 8.192 mS.
;               Increment the switch pressed counters if the corresponding PIC pin
;               is pulled low, else set the switch counter back to 0.
;               If any switches are pressed, set the display timeout down counter.
;
;               EVERY 8.192 mS.
;               Increment the 8 bit EEPROM write counter. When it wraps back to 0, set
;               the EEPROM write flag. (Indicates when to update the EEPROM, ie 2 seconds
;               after stopping the rotary encoder write any changes to EEPROM)
;
;               EVERY 8.192 mS.
;               Decrement the 16 bit display counter. when it reaches 0, set the
;               TIMEOUT flag.
;               (If the rotary encoder or buttons have not been changed for about 10)
;               (seconds, the flag indicates that its time to return from the setup )
;               (screens, back to the main display.                                 )
;
;*****************************************************************************************

                ORG     INTVECTOR

;-----------------------------------------------------------------------------------------
;
;               The following is processed every 1.024 mS.
;
;-----------------------------------------------------------------------------------------

                MOVWF   SAVE_W_REG              ; Save W reg & STATUS regs.
                SWAPF   STATUS,W
                MOVWF   SAVE_STATUS
                BCF     RP0                     ; Make sure we are addressing bank 0.

;-----------------------------------------------------------------------------------------
;
;               Keep track of the last 4 samples of the rotary encoder. (debouncing)
;
;-----------------------------------------------------------------------------------------

                RLF     ENC_HISTORY,F           ; Make room in the shift register for
                RLF     ENC_HISTORY,F           ; the two new encoder samples.
                MOVLW   B'11111100'
                ANDWF   ENC_HISTORY,F

                MOVLW   B'00000011'
                ANDWF   PORTA,W                 ; Get the current encoder state.
                IORWF   ENC_HISTORY,F           ; Put the new bits into the shift reg.

                MOVLW   B'01010101'
                ANDWF   ENC_HISTORY,W           ; Mask the last 4 encoder A samples.
                BTFSC   ZERO                    ; Were all 4 samples low ?
                BCF     ENC_A_DEBOUNCED         ; Y. Enc A is truly low !

                XORLW   B'01010101'             ; Mask the last 4 encoder A samples.
                BTFSC   ZERO                    ; Were all 4 samples high ?
                BSF     ENC_A_DEBOUNCED         ; Y. Enc A is truly high !

                MOVLW   B'10101010'
                ANDWF   ENC_HISTORY,W           ; Mask the last 4 encoder B samples.
                BTFSC   ZERO                    ; Were all 4 samples low ?
                BCF     ENC_B_DEBOUNCED         ; Y. Enc B is truly low !

                XORLW   B'10101010'             ; Mask the last 4 encoder B samples.
                BTFSC   ZERO                    ; Were all 4 samples high ?
                BSF     ENC_B_DEBOUNCED         ; Y. Enc B is truly high !

;-----------------------------------------------------------------------------------------
;
;               Determine if the rotary encoder has been moved.
;
;-----------------------------------------------------------------------------------------

                BTFSC   OLD_A_DEBOUNCED         ; Was old Enc A low ? (looking for +edge)
                GOTO    DECODE_DONE
                BTFSS   ENC_A_DEBOUNCED         ; Y. Is current encoder A high ?
                GOTO    DECODE_DONE
                BTFSS   ENC_B_DEBOUNCED         ;    Y. Is encoder B high ?
                GOTO    DEC_ENC_COUNT
INC_ENC_COUNT   MOVF    STEP_RATE,W             ;       N. Set ENC_COUNTER to 1 to 30.
                GOTO    SET_RATE                ;          (Slow CW rotation to fast CW)
DEC_ENC_COUNT   COMF    STEP_RATE,W             ;       Y. Set ENC_COUNTER to -1 to -30.
                ADDLW   1                       ;          (slow CCW rotation to fast CCW)
SET_RATE:       MOVWF   ENC_COUNTER
                MOVLW   38                      ; Set the step rate to 38. (dec every mS)
                MOVWF   STEP_RATE               ; (Max rate = 38 - (2 x 4mS debounce)=30)
                CALL    SET_TIMEOUT             ; Reset the display timeout down counter.
                CLRF    EEPROM_WRITE            ; Reset the EEPROM write counter.
DECODE_DONE:    BCF     OLD_A_DEBOUNCED         ; Clear the old debounced encoder bit.
                BTFSC   ENC_A_DEBOUNCED         ; Get the current debounced bit.
                BSF     OLD_A_DEBOUNCED         ; Save it for next time.
                DECFSZ  STEP_RATE,W             ; Dec the step rate counter,
                MOVWF   STEP_RATE               ; Do not dec below 1.


                DECFSZ  INT_COUNT,F             ; Dec the interrupt counter. Equal to 0 ?
                GOTO    RS_INT_STATUS           ; N. Exit the interrupt service routine.
                BSF     INT_COUNT,3             ; Y. Set the count back to 8. Continue
                                                ;    with the rest of the int routine.

;-----------------------------------------------------------------------------------------
;
;               The following is processed every 8.192 mS.
;
;-----------------------------------------------------------------------------------------

;-----------------------------------------------------------------------------------------
;               If CAL or Step size is pressed, reset the display timeout.
;-----------------------------------------------------------------------------------------

                BTFSS   SWITCH_A                ; Is the step size button pressed ?
                CALL    SET_TIMEOUT             ; Y. Set the display timeout down counter.
                BTFSS   SWITCH_B                ; Is the CAL button pressed ?
                CALL    SET_TIMEOUT             ; Y. Set the display timeout down counter.

;-----------------------------------------------------------------------------------------
;               Debounce the RIT button.
;-----------------------------------------------------------------------------------------

                INCFSZ  SW_BC_CNT,W             ; Inc the RIT button count.
                MOVWF   SW_BC_CNT               ; But dont inc past 255.
                MOVF    PORTA,W
                ANDLW   B'00011000'
                BTFSC   ZERO                    ; Are switch B & C pins low,
                GOTO    SW_END                  ; Y. Dont update CAL and TX buttons.
                CLRF    SW_BC_CNT               ; N. Clear the count.

;-----------------------------------------------------------------------------------------
;               Debounce the CAL button.
;-----------------------------------------------------------------------------------------

                INCFSZ  SW_B_CNT,W              ; Inc the CAL button count.
                MOVWF   SW_B_CNT                ; But dont inc past 255.
                BTFSC   SWITCH_B                ; Is the switch B pin low ?
                CLRF    SW_B_CNT                ; N. Clear the count.

;-----------------------------------------------------------------------------------------
;               Debounce the TX button.
;-----------------------------------------------------------------------------------------

                INCFSZ  SW_C_CNT,W              ; Inc the TX button count.
                MOVWF   SW_C_CNT                ; But dont inc past 255.
                BTFSC   SWITCH_C                ; Is the switch C pin low ?
                CLRF    SW_C_CNT                ; N. Clear the count.
SW_END:

;-----------------------------------------------------------------------------------------
;               Increment the EEPROM update counter.
;-----------------------------------------------------------------------------------------

                INCFSZ  EEPROM_WRITE,F          ; Inc the counter. Count = 0 ?
                GOTO    WR_DONE                 ; N. All done.
                BSF     EEPROM_UPDATE           ; Y. Set the flag. (update the EEPROM)
WR_DONE:

;-----------------------------------------------------------------------------------------
;               Decrement the Display timeout counter.
;-----------------------------------------------------------------------------------------

                DECFSZ  TIMEOUT_L,F             ; Dec the timeout counter.
                GOTO    T_DONE
                DECFSZ  TIMEOUT_H,F             ; Is it = 0 ?
                GOTO    T_DONE                  ; N. All done.
                BSF     TIMEOUT                 ; Y. Set the timeout flag.
T_DONE:

;-----------------------------------------------------------------------------------------
;
;               Return from interrupt.
;
;-----------------------------------------------------------------------------------------

RS_INT_STATUS:  SWAPF   SAVE_STATUS,W
                MOVWF   STATUS                  ; Restore STATUS reg.
                SWAPF   SAVE_W_REG,F            ; Save W reg.
                SWAPF   SAVE_W_REG,W            ; Restore W reg.
                BCF     T0IF                    ; Clear TMR0 interrupt flag.
                RETFIE

;-----------------------------------------------------------------------------------------
;
;               Reset the display timeout to 10 sec.
;
;-----------------------------------------------------------------------------------------

SET_TIMEOUT:    MOVLW   TIMEOUT_NUM             ; Load timeout with 10 sec.
                MOVWF   TIMEOUT_H
                CLRF    TIMEOUT_L
                RETURN

;*****************************************************************************************
;               End of interrupt service routine
;*****************************************************************************************


;*****************************************************************************************
;
;       Text table      (RETLW 'X')
;
;       Placed in page 0, so we dont have to bother with page bits.
;
;*****************************************************************************************

TABLE:          ADDWF   PCL,F                   ; Jump to character pointed to in W reg.
START:                                          ; Label the start of text strings.

RIT_OFF:        EQU $ - START
                DT              "RIT "
OFFSET:         EQU $ - START
                DT              "OFFSET ",0

TX:             EQU $ - START
                DT              "TX ",0

MIN_RX:         EQU $ - START
                DT              "MIN "
RX:             EQU $ - START
                DT              "RX ",0

MAX:            EQU $ - START
                DT              "MAX ",0

DDS_FREQ:       EQU $ - START
                DT              "DDS "
FREQ:           EQU $ - START
                DT              "FREQ ",0

MULTIP:         EQU $ - START
                DT              "MULTIPLIER",0

CAL:            EQU $ - START
                DT              "10"
MHZ:            EQU $ - START
                DT              " MHz CAL",0

VER:            EQU $ - START
                DT              " DD_SYNTH v1.10",0

TABLE_END:

;*****************************************************************************************
;
;               Initialise the hardware. (continued)
;
;*****************************************************************************************

CONTINUE:       CLRF    PORTB                   ; RB0..RB7 = Output.
                MOVLW   B'00000001'             ; Set prescaler to TMR0.
                MOVWF   OPTION_REG              ; TMR0 rate = 1:4 of instruction clock.
                BCF     RP0                     ; Return to page 0 for port access.

                MOVLW   B'10100000'             ; Enable TMR0 overflow interrupt.
                MOVWF   INTCON                  ; Sets interrupt rate to 1.024 mS.

                CLRF    PORTB                   ; Set all pins to LCD & DDS low.

                MOVLW   100
                CALL    MS_WAIT                 ; Wait until the LCD has powered up.

                MOVLW   COUNT_2                 ; Point to start of RAM we want to clear.
                MOVWF   FSR
                MOVLW   68-4                    ; Number of bytes of RAM to clear.
                CALL    CLEAR_BYTES             ; Clear the RAM.

                CALL    INIT_LCD                ; Initialise the LCD display.
                CALL    SEND_DDS_WORD           ; Initialise the DDS chip.

                MOVLW   VER
                CALL    LCD_TEXT                ; Display version info.
VER_LOOP:       BTFSS   EEPROM_WRITE,7          ; Wait 1 sec. (until EEPROM_WRITE > 127)
                GOTO    VER_LOOP

                BTFSS   SWITCH_B                ; Is the CAL button pressed.
                GOTO    CAL_10MHZ               ; Y. Display the calibration screens.

;*****************************************************************************************
;
;       Main display screen.
;
;               Get the frequency and step size from EEPROM.
;               Display RX or TX based on the state of the TX lead.
;               Continuously check the rotary encoder for movement, make sure the
;               frequency stays within set limits. If TX is operated, add the TX offset.
;               Then update the DDS chip if required. Add the frequency offset and
;               multiply the result by MULTIPLIER then display the result on the LCD.
;               If the step size button is pressed, allow the user to alter the step size.
;               If the frequency or step size have been changed update the EEPROM 
;               2 seconds after the rotary encoder stops moving.
;               If the CAL button has been pressed, display the setup screens.
;               If the RIT button has been pressed, display the RIT screen.
;
;*****************************************************************************************

MAIN_DISP:      CALL    LCD_CLEAR               ; Clear the display.
                CALL    EE_TO_RAM               ; Get frequency & step size from EEPROM.

MAIN_LOOP:      CALL    DISP_TX_RX              ; Display TX or RX on the LCD.
                CALL    UP_DOWN_RX              ; Inc or dec RX freq & adjust step size.

                BTFSC   EEPROM_UPDATE           ; Is the EEPROM update flag set?
                CALL    RAM_TO_EE               ; Y. Copy freq and step size to EEPROM.

                CALL    CHK_RIT_BRIEF
                BTFSC   PRESSED                 ; RIT button pressed briefly ?
                GOTO    RIT_DISP                ; Y. Goto the RIT display.

                CALL    CHK_TX
                BTFSC   PRESSED                 ; TX button pressed ?
                CALL    ADD_TX_OFFSET           ; Y. Add the TX offset.

                CALL    FREQ_TO_LCD             ; Update DDS, add offset and display it.

                BTFSS   SW_B_CNT,7              ; CAL button pressed for 1 sec ?
                GOTO    MAIN_LOOP               ; N. Continue updating the display.
                CALL    RAM_TO_EE               ; Y. Copy freq and step size to EEPROM.
                GOTO    SETUP_SCRNS             ;    Goto setup screens.

;*****************************************************************************************
;
;       RIT display screen.
;
;               Display RIT on the LCD.
;               Display RX or TX based on the state of the TX lead.
;               If TX is operated, add the TX offset to the RX frequency saved in EEPROM.
;               (not the RIT frequency)
;               If TX is not operated, continuously check the rotary encoder for
;               movement, make sure the RIT frequency stays within the RIT limits.
;               (RX frequency +/- RIT limit)
;               Then update the DDS chip if required. Add the frequency offset and
;               multiply the result by MULTIPLIER then display the result on the LCD.
;               If the step size button is pressed, allow the user to alter the step size.
;               Do not save any changes to EEPROM.
;               If the CAL button has been pressed, do nothing.
;               If the RIT button has been pressed, return to the main display.
;               (the RIT frequency is discarded upon return)
;
;*****************************************************************************************

RIT_DISP:       CALL    RAM_TO_EE               ; Copy frequency and step size to EEPROM.

                CLRF    STEP_NUM                ; Use the minimum step size, 1 Hz.
                BSF     RIT_SCRN                ; Set the RIT screen flag.
                CALL    LCD_CLEAR               ; Clear the display.

RIT_LOOP:       CALL    DISP_RIT                ; Display 'RIT' on line 1 pos 14, Also
                                                ; display TX or RX on the LCD.
                CALL    CHK_TX
                BTFSS   PRESSED                 ; TX button pressed ?
                GOTO    NOT_RIT_TX

                MOVLW   EE_RX_DDS_FREQ
                CALL    EE_TO_ARG2              ; Y. Get the RX DDS freq from EEPROM.

                CALL    ADD_TX_OFFSET           ;    Add the TX offset.

                CLRF    ENC_COUNTER             ;    Reset the encoder position.
                GOTO    RIT_DDS_UPDATE          ;    (dont use position info)

NOT_RIT_TX:     CALL    UP_DOWN_RX              ; N. Inc, dec RX freq & adjust step size.
                CALL    LIMIT_RIT               ;    Must be within the RIT range.

RIT_DDS_UPDATE: CALL    FREQ_TO_LCD             ; Update DDS, add offset and display it.

                CALL    CHK_RIT_BRIEF
                BTFSS   PRESSED                 ; RIT button pressed briefly ?
                GOTO    RIT_LOOP                ; N. Continue updating the RIT display.
                BCF     RIT_SCRN                ; Y. Clear the RIT screen flag.
                GOTO    MAIN_DISP               ;    Return to the main display.

;*****************************************************************************************
;
;       Calibration screen :- Adjust calibration DDS_CAL to give a DDS O/P of 10 MHz.
;                             (DDS_CAL) = 2^56 / (XTAL_FREQ)
;
;       Range = (EE_CAL_MIN) to (EE_CAL_MAX). Suitable for XTAL freq of 30 MHz to 130 MHz.
;
;*****************************************************************************************

CAL_10MHZ:      BSF     CAL_SCRN                ; Set the CAL screen flag.
                BSF     SW_B_CNT,7              ; Indicate CAL button pressed a long time.

                CALL    LCD_CLEAR               ; Clear the display.
                MOVLW   H'83'
                CALL    LCD_CMD                 ; Position the cursor, line 1 pos 4.
                MOVLW   CAL
                CALL    LCD_TEXT                ; Display '10 MHz CAL'.

                MOVLW   EE_DDS_CAL              ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

DDS_CAL:        CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                MOVLW   EE_CAL_MIN              ; Get the lower limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVLW   EE_CAL_MAX              ; Get the upper limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                MOVLW   EE_10MHZ                ; Get a 10 MHz constant.
                CALL    EE_TO_ARG3

                CALL    CHK_DDS                 ; Update DDS if there were any changes.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    DDS_CAL                 ; N. Continue updating the display.

                MOVLW   EE_DDS_CAL
                CALL    ARG2_TO_EE              ; Y. Save the CAL constant to EEPROM.
                BCF     CAL_SCRN                ;    Clear the CAL screen flag.

;*****************************************************************************************
;
;       Setup screen :- Maximum DDS frequency. (MAX_DDS_FREQ)
;
;       Range = 0 MHz to 60 MHz (Typically set to 1/3 of the DDS XTAL frequency)
;
;*****************************************************************************************

                CALL    LCD_HOME                ; Position the cursor, line 1 pos 1.
                MOVLW   MAX
                CALL    LCD_TEXT                ; Display 'MAX '.
                MOVLW   DDS_FREQ
                CALL    LCD_TEXT                ; Display 'DDS FREQ '.

                MOVLW   EE_MAX_DDS_FREQ         ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

                BCF     TIMEOUT                 ; Clear the display timeout flag.
MAX_DDS:        BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                CALL    CLEAR_ARG3              ; lower limit is 0.

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVLW   H'03'                   ; Set ARG3 to 60 MHz.
                MOVWF   ARG3_3                  ; ARG3 = 0003938700 HEX
                MOVLW   H'93'
                MOVWF   ARG3_2
                MOVLW   H'87'
                MOVWF   ARG3_1

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    MAX_DDS                 ; N. Continue updating the display.

                MOVLW   EE_MAX_DDS_FREQ
                CALL    ARG2_TO_EE              ; Y. Save the maximum DDS freq to EEPROM.

;*****************************************************************************************
;
;       Setup screen :- Minimum RX DDS frequency. (MIN_RX_DDS_FREQ)
;
;       Range = -(MAX_DDS_FREQ) to (MAX_DDS_FREQ)
;
;*****************************************************************************************

SETUP_SCRNS:    CALL    LCD_HOME                ; Position the cursor, line 1 pos 1.
                MOVLW   MIN_RX
                CALL    LCD_TEXT                ; Display 'MIN RX '.
                MOVLW   DDS_FREQ
                CALL    LCD_TEXT                ; Display 'DDS FREQ'.

                MOVLW   EE_MIN_RX_DDS_FREQ      ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

                BCF     TIMEOUT                 ; Clear the display timeout flag.
MIN_RX_DDS:     BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)

                MOVLW   EE_MAX_DDS_FREQ         ; Get the upper limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    NEG_ARG3                ; Set the limit to negative.

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    MIN_RX_DDS              ; N. Continue updating the display.

                MOVLW   EE_MIN_RX_DDS_FREQ
                CALL    ARG2_TO_EE              ; Y. Save the minimum RX freq to EEPROM.

                BCF     NEGATIVE_RX
                BTFSC   ARG2_4,7                ; Was the result negative ?
                BSF     NEGATIVE_RX             ; Y. Save it for future use.

;*****************************************************************************************
;
;       Setup screen :- Maximum RX DDS frequency. (MAX_RX_DDS_FREQ)
;
;       If (MIN_RX_DDS_FREQ) is positive.
;       Range = (MIN_RX_DDS_FREQ) to (MAX_DDS_FREQ)
;
;       If (MIN_RX_DDS_FREQ) is negative,
;       Range = (MIN_RX_DDS_FREQ) to 0
;
;*****************************************************************************************

                CALL    LCD_HOME                ; Position the cursor, line 1 pos 1.
                MOVLW   MAX
                CALL    LCD_TEXT                ; Display 'MAX '.

                MOVLW   EE_MAX_RX_DDS_FREQ      ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

MAX_RX_DDS:     BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                MOVLW   EE_MIN_RX_DDS_FREQ      ; Get the lower limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVLW   EE_MAX_DDS_FREQ         ; Prepare the upper limit.
                CALL    EE_TO_ARG3

                BTFSC   NEGATIVE_RX             ; Was the lower RX limit negative ?
                CALL    CLEAR_ARG3              ; Y. Change the upper limit to 0.

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    MAX_RX_DDS              ; N. Continue updating the display.

                MOVLW   EE_MAX_RX_DDS_FREQ
                CALL    ARG2_TO_EE              ; Y. Save the maximum RX freq to EEPROM.

;*****************************************************************************************
;
;       Setup screen :- Maximum RIT offset frequency. (MAX_RIT_OFFSET)
;
;       Range = 0 to (MAX_DDS_FREQ)
;
;*****************************************************************************************

                MOVLW   H'84'                   ; Position the cursor, line 1 pos 5.
                CALL    LCD_CMD
                MOVLW   RIT_OFF
                CALL    LCD_TEXT                ; Display 'RIT OFFSET'.

                MOVLW   EE_MAX_RIT_OFFSET       ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

RIT_OFFSET:     BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                CALL    CLEAR_ARG3              ; Lower limit of 0.

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVLW   EE_MAX_DDS_FREQ         ; Get the upper limit.
                CALL    EE_TO_ARG3

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    RIT_OFFSET              ; N. Continue updating the display.

                MOVLW   EE_MAX_RIT_OFFSET
                CALL    ARG2_TO_EE              ; Y. Save the maximum RX freq to EEPROM.

;*****************************************************************************************
;
;       Setup screen :- TX offset frequency. (TX_OFFSET_FREQ)
;
;       If (MIN_RX_DDS_FREQ) is positive,
;       Range = 0-(MIN_RX_DDS_FREQ) to (MAX_DDS_FREQ)-(MAX_RX_DDS_FREQ)
;
;       If (MIN_RX_DDS_FREQ) is negative,
;       Range = -(MAX_DDS_FREQ)-(MIN_RX_DDS_FREQ) to 0-(MAX_RX_DDS_FREQ)
;
;*****************************************************************************************

                CALL    LCD_HOME                ; Position the cursor, line 1 pos 1.
                MOVLW   TX
                CALL    LCD_TEXT                ; Display 'TX '.
                MOVLW   OFFSET
                CALL    LCD_TEXT                ; Display 'OFFSET '.
                MOVLW   FREQ
                CALL    LCD_TEXT                ; Display 'FREQ '.

                MOVLW   EE_TX_OFFSET_FREQ       ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

TX_OFFSET:      BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                CALL    PUSH_ARG2               ; Save the variable we are adjusting.

                MOVLW   EE_MAX_DDS_FREQ         ; Prepare the lower limit.
                CALL    EE_TO_ARG2

                CALL    NEG_ARG2                ; Make the it negative.

                BTFSS   NEGATIVE_RX             ; Was the lower RX limit negative ?
                CALL    CLEAR_ARG2              ; N. Change the limit to 0.

                MOVLW   EE_MIN_RX_DDS_FREQ      ; Get the lower limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    SUB                     ; Calculate the lower limit.

                CALL    PULL_ARG3               ; Recover the variable we are adjusting.

                CALL    MAXIMUM                 ; Make sure we dont go below the limit.

                CALL    PUSH_ARG2               ; Save the variable we are adjusting.

                MOVLW   EE_MAX_DDS_FREQ         ; Prepare the upper limit.
                CALL    EE_TO_ARG2

                BTFSC   NEGATIVE_RX             ; Was the lower RX limit negative ?
                CALL    CLEAR_ARG2              ; Y. Change the limit to 0.

                MOVLW   EE_MAX_RX_DDS_FREQ      ; Get the upper limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    SUB                     ; Calculate the upper limit.

                CALL    PULL_ARG3               ; Recover the variable we are adjusting.

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    TX_OFFSET               ; N. Continue updating the display.

                MOVLW   EE_TX_OFFSET_FREQ
                CALL    ARG2_TO_EE              ; Y. Save the TX offset freq to EEPROM.

;*****************************************************************************************
;
;       Setup screen :- Offset frequency. (OFFSET_FREQ)
;
;       Range = -xx MHz to 20 GHz.
;       (allow negative offset, resulting display must always be positive)
;       (Lower limit = -(MINIMUM(MIN_RX_DDS_FREQ+TX_OFFSET_FREQ),(MIN_RX_DDS_FREQ))
;
;*****************************************************************************************

                CALL    LCD_CLEAR               ; Clear LCD & set cursor, line 1 pos 1.
                MOVLW   OFFSET
                CALL    LCD_TEXT                ; Display 'OFFSET '.
                MOVLW   FREQ
                CALL    LCD_TEXT                ; Display 'FREQ '.

                MOVLW   EE_OFFSET_FREQ          ; Get the variable we want to adjust
                CALL    EE_TO_ARG2              ; from EEPROM.

OFFSET_FREQU:   BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                CALL    PUSH_ARG2               ; Save the variable we are adjusting.

                MOVLW   EE_TX_OFFSET_FREQ       ; Prepare the lower limit.
                CALL    EE_TO_ARG2

                MOVLW   EE_MIN_RX_DDS_FREQ      ; Prepare the lower limit.
                CALL    EE_TO_ARG3

                CALL    ADD                     ; Calculate the lower limit.
                CALL    MINIMUM
                CALL    NEG_ARG2

                CALL    PULL_ARG3               ; Recover the variable we are adjusting.

                CALL    MAXIMUM                 ; Make sure we dont go below the limit.

                MOVLW   EE_MAX_OFFSET           ; Get the upper limit from EEPROM.
                CALL    EE_TO_ARG3

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    DISPLAY_FREQ            ; Display the variable.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    OFFSET_FREQU            ; N. Continue updating the display.

                MOVLW   EE_OFFSET_FREQ
                CALL    ARG2_TO_EE              ; Y. Save the offset freq to EEPROM.

;*****************************************************************************************
;
;       Setup screen :- Frequency display multiplier. (DISPLAY_MULT)
;
;       Range = 1 to 20 (as long as any display on LCD is not to far above 20 GHz)
;
;*****************************************************************************************

                CLRF    STEP_NUM                ; Use the minimum step size. (1)
                MOVLW   20
                MOVWF   COUNT_2
                CLRF    COUNT_4

                MOVLW   EE_MAX_OFFSET           ; Get the max display freq from EEPROM.
                CALL    EE_TO_ARG2

                MOVLW   EE_OFFSET_FREQ          ; Get the offset freq from EEPROM.
                CALL    EE_TO_ARG3

MULT_DISP_LOOP: INCF    COUNT_4,F
                CALL    SUB                     ; ARG2 = ARG2 - offset freq.
                BTFSC   ARG2_4,7                ; Is the result negative ?
                GOTO    GOT_LIMIT               ; Y. Done.
                DECFSZ  COUNT_2,F               ; N. Continue subtracting. (Max 20 times)
                GOTO    MULT_DISP_LOOP
                                                ; COUNT_4 = upper limit.
GOT_LIMIT:      CALL    LCD_CLEAR               ; Clear LCD & set cursor, line 1 pos 1.
                MOVLW   MULTIP
                CALL    LCD_TEXT                ; Display 'MULTIPLIER'.

                CALL    CLEAR_ARG2              ; Move the mult byte from EEPROM to ARG2.
                CALL    READ_MULT
                MOVWF   ARG2_0

MULT_DISP:      BTFSC   TIMEOUT                 ; Has the display timeout been reached ?
                GOTO    MAIN_DISP               ; Y. Return to the main display.

                CALL    UP_DOWN_A               ; Chk rotary encoder. (inc or dec ARG2)

                CALL    CLEAR_ARG3              ; Lower limit is 1.
                CALL    INC_ARG3

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVF    COUNT_4,W
                MOVWF   ARG3_0                  ; Get the upper limit.

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                CALL    BIN_TO_DEC              ; Convert the number to ASCII decimal.

                MOVLW   H'8E'                   ; Position the cursor, line 1 pos 15.
                CALL    LCD_CMD

                MOVF    ARG1_1,W                ; Display the number in ASCII form.
                CALL    LCD_CHR                 ; 10's digit.
                MOVF    ARG1_0,W
                CALL    LCD_CHR                 ; 1's digit.

                CALL    CHK_CAL_BRIEF
                BTFSS   PRESSED                 ; CAL button pressed briefly ?
                GOTO    MULT_DISP               ; N. Continue updating the display.
                CALL    WRITE_MULT              ; Y. Save the multiplier to EEPROM.
                GOTO    MAIN_DISP               ;    Return to the main display.

;*****************************************************************************************
;               Start of subroutines
;*****************************************************************************************



;*****************************************************************************************
;
;      NAME:    ADD_TX_OFFSET
;
;   PURPOSE:    Add the TX offset (from EEPROM) to ARG2.
;
;     INPUT:    ARG2 = frequency.
;
;    OUTPUT:    ARG2 = frequency + TX offset.
;               (ARG3 destroyed)
;
;*****************************************************************************************

ADD_TX_OFFSET:  MOVLW   EE_TX_OFFSET_FREQ       ; Get the TX offset.
                CALL    EE_TO_ARG3              ; Fall through to ADD,
                                                ; and return via ADD.

;*****************************************************************************************
;
;     NAMES:    ADD                     ARG2 = ARG2 + ARG3.     (all routines are 40 bit)
;               SUB                     ARG2 = ARG2 - ARG3.
;               INC_ARG2                ARG2 = ARG2 + 1.
;               INC_ARG3                ARG3 = ARG3 + 1.
;               NEG_ARG2                ARG2 = -ARG2.
;               NEG_ARG2                ARG2 = -ARG2.
;               CLEAR_ARG1              ARG1 = 0.
;               CLEAR_ARG2              ARG2 = 0.
;               CLEAR_ARG3              ARG3 = 0.
;
;*****************************************************************************************

ADD:            MOVF    ARG3_0,W
                ADDWF   ARG2_0,F                ; Add the first (LSD) bytes.
                MOVF    ARG3_1,W
                BTFSC   CARRY
                INCFSZ  ARG3_1,W                ; If there was a carry, inc the next byte.

                ADDWF   ARG2_1,F                ; Add the second bytes.
                MOVF    ARG3_2,W
                BTFSC   CARRY
                INCFSZ  ARG3_2,W                ; If there was a carry, inc the next byte.

                ADDWF   ARG2_2,F                ; Add the third bytes.
                MOVF    ARG3_3,W
                BTFSC   CARRY
                INCFSZ  ARG3_3,W                ; If there was a carry, inc the next byte.

                ADDWF   ARG2_3,F                ; Add the fourth bytes.
                MOVF    ARG3_4,W
                BTFSC   CARRY
                INCFSZ  ARG3_4,W                ; If there was a carry, inc the next byte.

                ADDWF   ARG2_4,F                ; Add the fith bytes.
                RETURN


SUB:            CALL    NEG_ARG3                ; ARG3 = -ARG3.
                CALL    ADD                     ; ARG2 = ARG2 + -ARG3.
NEG_ARG3:       COMF    ARG3_0,F                ;   (negate ARG3)
                COMF    ARG3_1,F
                COMF    ARG3_2,F
                COMF    ARG3_3,F
                COMF    ARG3_4,F
INC_ARG3:       INCFSZ  ARG3_0,F                ;   (add 1 to ARG3)
                RETURN
                INCFSZ  ARG3_1,F
                RETURN
                INCFSZ  ARG3_2,F
                RETURN
                INCFSZ  ARG3_3,F
                RETURN
                INCF    ARG3_4,F
                RETURN

NEG_ARG2:       COMF    ARG2_0,F                ; Negate ARG2.
                COMF    ARG2_1,F
                COMF    ARG2_2,F
                COMF    ARG2_3,F
                COMF    ARG2_4,F
INC_ARG2:       INCFSZ  ARG2_0,F                ; Add 1 to ARG2.
                RETURN
                INCFSZ  ARG2_1,F
                RETURN
                INCFSZ  ARG2_2,F
                RETURN
                INCFSZ  ARG2_3,F
                RETURN
                INCF    ARG2_4,F
                RETURN


CLEAR_ARG1:     CALL    SET_ARG1                ; Set FSR = ARG1_10, COUNT_1 = 11.
                GOTO    CLEAR_LOOP

CLEAR_ARG2:     MOVLW   ARG2_4
                GOTO    CLEAR

CLEAR_ARG3:     MOVLW   ARG3_4
CLEAR:          MOVWF   FSR
                MOVLW   5
CLEAR_BYTES:    MOVWF   COUNT_1                 ; Save the count.
CLEAR_LOOP:     CLRF    INDF                    ; Clear the RAM byte.
                INCF    FSR,F                   ; Inc the RAM address pointer.
                DECFSZ  COUNT_1,F               ; Dec the count.
                GOTO    CLEAR_LOOP              ; Loop until all bytes cleared.
                RETURN

SET_ARG1:       MOVLW   ARG1_10                 ; Set FSR = ARG1_10, COUNT_1 = 11.
                MOVWF   FSR
                MOVLW   11
                MOVWF   COUNT_1
                RETURN

;*****************************************************************************************
;
;      NAME:    CMP
;
;   PURPOSE:    40 bit compare.  ARG2 - ARG3
;
;     INPUT:    40 bit argment ARG2.
;               40 bit argment ARG3.
;
;    OUTPUT:    None.
;               Carry flag clear (borrow) if result was negative.
;               Zero flag set if equal.
;               (ARG2 & ARG3 unchanged)
;
;*****************************************************************************************

CMP:            CALL    OFFSET_ARGS             ; Add a large offset to ARG2 & ARG3.
                                                ; (makes sure we are dealing with)
                                                ; (positive numbers only         )
                MOVF    ARG3_0,W
                SUBWF   ARG2_0,W                ; Sub the first (LSD) bytes.
                MOVWF   TEMP_1                  ; Save the result for zero check.
                MOVF    ARG3_1,W
                BTFSS   CARRY
                INCFSZ  ARG3_1,W                ; If there was a carry, inc the next byte.

                SUBWF   ARG2_1,W                ; Sub the second bytes.
                IORWF   TEMP_1,F                ; Save the result for zero check.
                MOVF    ARG3_2,W
                BTFSS   CARRY
                INCFSZ  ARG3_2,W                ; If there was a carry, inc the next byte.

                SUBWF   ARG2_2,W                ; Sub the third bytes.
                IORWF   TEMP_1,F                ; Save the result for zero check.
                MOVF    ARG3_3,W
                BTFSS   CARRY
                INCFSZ  ARG3_3,W                ; If there was a carry, inc the next byte.

                SUBWF   ARG2_3,W                ; Sub the fourth bytes.
                IORWF   TEMP_1,F                ; Save the result for zero check.
                MOVF    ARG3_4,W
                BTFSS   CARRY
                INCFSZ  ARG3_4,W                ; If there was a carry, inc the next byte.

                SUBWF   ARG2_4,W                ; Sub the fith bytes.
                IORWF   TEMP_1,F                ; Save the result for zero check.

OFFSET_ARGS:    MOVLW   B'10000000'             ; Remove the offset from ARG2 and ARG3.
                XORWF   ARG2_4,F
                XORWF   ARG3_4,F
                MOVF    TEMP_1,W                ; Get the zero check byte.
                RETURN                          ; (sets zero flag if result was 0)

;*****************************************************************************************
;
;      NAME:    MINIMUM
;
;   PURPOSE:    40 bit minimum. Return the smallest of two argments.
;
;       INPUT:  40 bit argment in ARG2.
;               40 bit argment in ARG3.
;
;       OUTPUT: The smallest argment is returned in ARG2.
;               (ARG3 unchanged)
;
;*****************************************************************************************

MINIMUM:        CALL    CMP
                BTFSS   CARRY                   ; Is ARG2 > ARG3 ?
                RETURN                          ; N. No change required.
                GOTO    ARG3_TO_ARG2            ; Y. Set ARG2 the same as ARG3.
                                                ;    Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    MAXIMUM
;
;   PURPOSE:    40 bit maximum. Return the largest of two argments.
;
;       INPUT:  40 bit argment in ARG2.
;               40 bit argment in ARG3.
;
;       OUTPUT: The largest argment is returned in ARG2.
;               (ARG3 unchanged)
;
;*****************************************************************************************

MAXIMUM:        CALL    CMP
                BTFSC   CARRY                   ; Is ARG2 > ARG3 ?
                RETURN                          ; Y. No change required.
                GOTO    ARG3_TO_ARG2            ; N. Set ARG2 the same as ARG3.
                                                ;    Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    UP_DOWN_RX
;
;   PURPOSE:    If the rotary encoder has been moved, increase or decrease the
;               FREQUENCY variable by step size. Make sure it stays within limit.
;               Also see if the user wants to change the step size.
;
;     INPUT:    None. (freq in the FREQUENCY variable)
;
;    OUTPUT:    ARG2 = frequency +- N. (result also placed back in FREQUENCY variable)
;
;*****************************************************************************************

UP_DOWN_RX:     MOVLW   FREQUENCY
                CALL    COPY_TO_ARG2            ; Get the FREQUENCY variable.

                CALL    UP_DOWN                 ; Chk rotary encoder & adjust step size.
                                                ; (inc or dec ARG2)
                MOVLW   EE_MIN_RX_DDS_FREQ
                CALL    EE_TO_ARG3              ; Get the lower limit.

                CALL    MAXIMUM                 ; Make sure we dont go below that limit.

                MOVLW   EE_MAX_RX_DDS_FREQ
                CALL    EE_TO_ARG3              ; Get the upper limit.

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

ARG2_TO_FREQ:   MOVLW   FREQUENCY               ; Save the FREQUENCY variable.
                GOTO    COPY_FROM_ARG2          ; Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    LIMIT_RIT
;
;   PURPOSE:    When in the RIT display, make sure the RIT frequency (FREQUENCY variable)
;               stays within the limit set by EE_MAX_RIT_OFFSET.
;
;     INPUT:    None. (freq in the FREQUENCY variable)
;
;    OUTPUT:    ARG2 = frequency. (result also placed back in FREQUENCY variable)
;
;*****************************************************************************************

LIMIT_RIT:      MOVLW   EE_RX_DDS_FREQ
                CALL    EE_TO_ARG2              ; Prepare the upper limit.

                MOVLW   EE_MAX_RIT_OFFSET
                CALL    EE_TO_ARG3              ; Prepare the upper limit.

                CALL    ADD                     ; ARG2 = RX DDS freq + max RIT offset.

                CALL    PUSH_ARG2               ; Save the result. (upper limit)

                CALL    SUB                     ; ARG2 = RX DDS freq
                CALL    SUB                     ; ARG2 = RX DDS freq - max RIT offset.
                                                ;      = (lower limit)
                MOVLW   FREQUENCY               ;
                CALL    COPY_TO_ARG3            ; Get the FREQUENCY variable.

                CALL    MAXIMUM                 ; Make sure we dont go below the limit.

                CALL    PULL_ARG3               ; Get the upper limit.

                CALL    MINIMUM                 ; Make sure we dont go above that limit.

                GOTO    ARG2_TO_FREQ            ; Save the FREQUENCY variable.
                                                ; Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    UP_DOWN
;
;   PURPOSE:    If the rotary encoder has been moved, inc or dec the variable in ARG2.
;               Also see if the user wants to change the step size.
;
;     INPUT:    ARG2.
;
;    OUTPUT:    ARG2 = ARG2 +- N.
;
;*****************************************************************************************

UP_DOWN:        CALL    CHK_STEP_SIZE           ; Does the user want to change step size.
UP_DOWN_A:      MOVF    ENC_COUNTER,W           ; Any change to the rotary encoder ?
                BTFSC   ZERO                    ; N. Just return.
                RETURN

                MOVWF   COUNT_2                 ; Save the count.
                CLRF    ENC_COUNTER             ; Reset the encoder position count.

                CALL    GET_STEP_SIZE           ; Get the step size, place it in ARG3.
                BTFSS   COUNT_2,7               ; Is the encoder count negative ?
                GOTO    STEP_UP

STEP_DOWN:      CALL    SUB                     ; Y. Dec ARG2 by the step size.
                INCFSZ  COUNT_2,F               ;    Dec the loop count.
                GOTO    STEP_DOWN               ;    Dec ARG2 until count = 0.
                RETURN

STEP_UP:        CALL    ADD                     ; N. Inc ARG2 by the step size.
                DECFSZ  COUNT_2,F               ;    Dec the loop count.
                GOTO    STEP_UP                 ;    Inc ARG2 until count = 0.
                RETURN

;*****************************************************************************************
;
;      NAME:    GET_STEP_SIZE
;                                             (STEP_NUM)
;   PURPOSE:    Calculate the step size. = 10^
;
;     INPUT:    None.              (STEP_NUM = 0 to 7)
;
;    OUTPUT:    ARG3 = step size.  (1 ... 10,000,000)
;
;*****************************************************************************************

GET_STEP_SIZE:  CALL    PUSH_ARG2               ; Dont destroy the value in ARG2.
                CALL    CLEAR_ARG2
                CALL    INC_ARG2                ; Set ARG2 to 1.
                INCF    STEP_NUM,W              ; Get the step number + 1.
                MOVWF   COUNT_3
                GOTO    LOOP_CHK

TIMES_TEN:      CALL    ADD                     ; ARG2 = ARG3 x 10.
                DECFSZ  COUNT_0,F
                GOTO    TIMES_TEN

LOOP_CHK:       MOVLW   9
                MOVWF   COUNT_0                 ; Count for 10 times multiplication.
                CALL    ARG2_TO_ARG3            ; Move ARG2 (result) to ARG3.
                DECFSZ  COUNT_3,F               ; Keep mult by ten until counter = 0.
                GOTO    TIMES_TEN

                GOTO    PULL_ARG2               ; Return the original value of ARG2.
                                                ; Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    CHK_STEP_SIZE
;
;   PURPOSE:    If the user wants to change the step size, enable the cursor on the
;               frequency display, and set the STEP_NUM variable.
;
;     INPUT:    None.
;
;    OUTPUT:    None.
;
;*****************************************************************************************

CHK_STEP_SIZE:  BTFSC   SWITCH_A                ; Adjust step size button pressed ?
                RETURN                          ; N. All done.
                MOVLW   B'00001110'             ; Turn on the cursor.
                CALL    LCD_CMD
CHK_STEP_LOOP:  MOVLW   6
                BTFSC   CAL_SCRN                ;  (if we are in the CAL 10 MHz screen)
                MOVLW   8                       ;  (change the cursor position.       )
                SUBWF   STEP_NUM,W              ; Is cursor before the decimal point ?
                BTFSS   CAL_SCRN                ;  (if we are in the CAL 10 MHz screen)
                BTFSC   CARRY                   ;  (ignore the decimal point.         )
                ADDLW   1                       ; Y. Shift cursor left 1 digit.
                SUBLW   H'C5'                   ; Subtract it from the 1Hz char pos.
                CALL    LCD_CMD                 ; Position the cursor on freq display.

                MOVF    ENC_COUNTER,W           ; Get rotary encoder position.
                CLRF    ENC_COUNTER             ; Reset the encoder position.
                SUBWF   STEP_NUM,F              ; Take it from the step position.
                BTFSC   STEP_NUM,7              ; Is the result < 0 ?
                CLRF    STEP_NUM                ; Y. Limit it to 0.
                MOVLW   7
                SUBWF   STEP_NUM,W
                MOVLW   7
                BTFSC   CARRY                   ; Was the result > 7 ?
                MOVWF   STEP_NUM                ; Y. Limit it to 7.

                BTFSS   SWITCH_A                ; Adjust step size button still pressed ?
                GOTO    CHK_STEP_LOOP           ; Y. Keep checking rotary encoder.
                GOTO    LCD_NORM                ; N. Turn off the cursor.
                                                ;    Return via "GOTO".

;*****************************************************************************************
;
;     NAMES:    COPY_TO_ARG2            Copy 5 bytes pointed to by W to ARG2.
;               COPY_TO_ARG3            Copy 5 bytes pointed to by W to ARG3.
;               COPY_FROM_ARG2          Copy ARG2 to the address pointed to by W.
;               COPY_FROM_ARG3          Copy ARG3 to the address pointed to by W.
;               ARG2_TO_ARG3            Copy ARG2 to ARG3.
;               ARG3_TO_ARG2            Copy ARG3 to ARG2.
;               PUSH_ARG2               Copy ARG2 to ARG4.
;               PUSH_ARG3               Copy ARG3 to ARG4.
;               PULL_ARG2               Copy ARG4 to ARG2.
;               PULL_ARG3               copy ARG4 to ARG3.
;
;   PURPOSE:    Move 5 bytes of RAM from source address to dest address.
;               (starting at source and dest, then source+1 dest+1, etc.)
;
;               RAM usage       TEMP_1 = destination address.
;                               TEMP_2 = source address.
;                               TEMP_3 = byte to move.
;
;*****************************************************************************************

PUSH_ARG2:      MOVLW   ARG4_4                  ; \
COPY_FROM_ARG2: MOVWF   TEMP_1                  ;  \
                MOVLW   ARG2_4                  ;   \
                GOTO    COPY_B                  ;    |
PUSH_ARG3:      MOVLW   ARG4_4                  ;    |
COPY_FROM_ARG3: MOVWF   TEMP_1                  ;    |
                MOVLW   ARG3_4                  ;    |
COPY_B:         MOVWF   TEMP_2                  ;    |
                GOTO    COPY_A                  ;    |
ARG3_TO_ARG2:   MOVLW   ARG3_4                  ;    |
                GOTO    COPY_TO_ARG2            ;     \  set the various source and
PULL_ARG2:      MOVLW   ARG4_4                  ;     /  destination addresses.
COPY_TO_ARG2:   MOVWF   TEMP_2                  ;    |
                MOVLW   ARG2_4                  ;    |
                GOTO    COPY                    ;    |
ARG2_TO_ARG3:   MOVLW   ARG2_4                  ;    |
                GOTO    COPY_TO_ARG3            ;    |
PULL_ARG3:      MOVLW   ARG4_4                  ;    |
COPY_TO_ARG3:   MOVWF   TEMP_2                  ;   /
                MOVLW   ARG3_4                  ;  /
COPY:           MOVWF   TEMP_1                  ; /
COPY_A:         MOVLW   5                       ; 5 bytes to move.
                MOVWF   COUNT_1
COPY_LOOP:      MOVF    TEMP_2,W                ; Get source address.
                MOVWF   FSR                     ;        "
                MOVF    INDF,W                  ; Get the byte from the source address.
                MOVWF   TEMP_3                  ;                "
                MOVF    TEMP_1,W                ; Get destination address.
                MOVWF   FSR                     ;          "
                MOVF    TEMP_3,W                ; Place the byte at destination address.
                MOVWF   INDF                    ;                   "
                INCF    TEMP_1,F                ; Inc dest address.
                INCF    TEMP_2,F                ; Inc source address.
                DECFSZ  COUNT_1,F
                GOTO    COPY_LOOP               ; Loop until all bytes copied.
                RETURN

;*****************************************************************************************
;
;     NAMES:    EE_TO_ARG2              Copy 5 bytes in EEPROM pointed to by W to ARG2.
;               EE_TO_ARG3              Copy 5 bytes in EEPROM pointed to by W to ARG3.
;               EE_TO_RAM               Copy FREQUENCY (5 bytes) and STEP_SIZE (1 BYTE)
;                                       from EEPROM to RAM.
;               RAM_TO_EE               Copy FREQUENCY (5 bytes) and STEP_SIZE (1 BYTE)
;                                       from RAM to EEPROM.
;               ARG2_TO_EE              Copy ARG2 to the EEPROM address pointed to by W.
;
;*****************************************************************************************


RAM_TO_EE:      CALL    SET_ADDRESS             ; Set the RAM and EEPROM address.
                GOTO    W_EEPROM
ARG2_TO_EE:     CALL    ARG2_ADDRESS            ;              "
W_EEPROM        MOVWF   COUNT_1                 ; Save the number of bytes to program.
W_EEPROM_LOOP:  CALL    WRITE_EEPROM
                INCF    EEADR,F                 ; Prepare to read the next byte.
                INCF    FSR,F
                DECFSZ  COUNT_1,F
                GOTO    W_EEPROM_LOOP           ; Loop until all bytes are updated.
                BCF     EEPROM_UPDATE           ; Clear the update required flag.
                RETURN


EE_TO_RAM:      CALL    SET_ADDRESS             ; Set the RAM and EEPROM address.
                GOTO    R_EEPROM
EE_TO_ARG3:     MOVWF   EEADR                   ;              "
                MOVLW   ARG3_4
                CALL    ARG_ADDRESS
                GOTO    R_EEPROM
EE_TO_ARG2:     CALL    ARG2_ADDRESS            ;              "
R_EEPROM        MOVWF   COUNT_1                 ; Save the number of bytes to read.
R_EEPROM_LOOP:  CALL    READ_EEPROM             ; Get a byte from EEPROM.
                MOVWF   INDF                    ; Place it in RAM.
                INCF    EEADR,F                 ; Prepare to read the next byte.
                INCF    FSR,F
                DECFSZ  COUNT_1,F
                GOTO    R_EEPROM_LOOP           ; Loop until all bytes read.
                RETURN


ARG2_ADDRESS:   MOVWF   EEADR                   ; The address in EEPROM is in W.
                MOVLW   ARG2_4
ARG_ADDRESS:    MOVWF   FSR                     ; The address in RAM is ARG2.
                MOVLW   5                       ; 6 bytes to move.
                RETURN

SET_ADDRESS:    MOVLW   EE_RX_DDS_FREQ          ; The address in EEPROM is EE_RX_DDS_FREQ.
                MOVWF   EEADR
                MOVLW   FREQUENCY               ; The address in RAM is FREQUENCY.
                MOVWF   FSR
                MOVLW   6                       ; 6 bytes to move.
                RETURN

;*****************************************************************************************
;
;      NAME:    WRITE_EEPROM
;
;   PURPOSE:    Write the byte of RAM pointed to by FSR,
;               to the EEPROM address pointed to by EEADR.
;               If the RAM and EEPROM data are the same, the EEPROM is not reprogrammed.
;               Assumes the source address FSR has been set,
;               and the destination address EEADR has been set.
;
;     INPUT:    None.
;
;    OUTPUT:    None.
;
;*****************************************************************************************

WRITE_MULT:     MOVLW   EE_DISPLAY_MULT         ; Dest address is EE_DISPLAY_MULT.
                MOVWF   EEADR                   ; Save the dest address.
                MOVLW   ARG2_0                  ; Source address is ARG2.
                MOVWF   FSR                     ; Save it in the indirect register.

WRITE_EEPROM:   CALL    READ_EEPROM             ; Get the current byte from EEPROM.
                SUBWF   INDF,W
                BTFSC   ZERO                    ; Is it the same as it is in RAM ?
                GOTO    NO_CHANGE
                MOVF    INDF,W                  ; N. Get the data byte from RAM.
                MOVWF   EEDATA                  ;    Put it in the buffer.
                BSF     RP0                     ;    Select page 1 for TRIS reg access.
                BCF     EEIF                    ;    Clear the EEPROM write done flag.
                BSF     WREN                    ;    Enable EEPROM write.
                MOVLW   H'55'                   ;    Perform required safety steps.
                MOVWF   EECON2
                MOVLW   H'AA'
                MOVWF   EECON2
                BSF     WR                      ;    Begin the write.
WRITE_LOOP:     BTFSS   EEIF                    ;    Loop until the write is completed.
                GOTO    WRITE_LOOP
                BCF     RP0                     ;    Return to page 0 for port access.
NO_CHANGE:      RETURN                          ; Y. No need to update this byte.

;*****************************************************************************************
;
;      NAME:    READ_EEPROM
;
;   PURPOSE:    Read the byte of EEPROM data pointed to by EEADR.
;               Assumes the read address EEADR has been set.
;
;     INPUT:    None.
;
;    OUTPUT:    W = The byte read from EEPROM.
;
;*****************************************************************************************

READ_MULT:      MOVLW   EE_DISPLAY_MULT         ; Source address is EE_DISPLAY_MULT.
                MOVWF   EEADR                   ; Save the source address.

READ_EEPROM:    BSF     RP0                     ; Select page 1 for TRIS reg access.
                BSF     RD                      ; Perform an EEPROM read.
                BCF     RP0                     ; Return to page 0 for port access.
                MOVF    EEDATA,W                ; Get the read byte.
                RETURN

;*****************************************************************************************
;
;     NAMES:    CHK_CAL_BRIEF   Return a true flag in PRESSED, if the
;               CHK_RIT_BRIEF   button was pressed briefly.
;
;               CHK_TX          Return a true flag in PRESSED, if the
;                               button has been pressed for a while.
;     INPUT:    None.
;
;    OUTPUT:    None. (PRESSSED flag set if button pressed)
;
;*****************************************************************************************

CHK_TX          MOVLW   MIN_ON                  ; Use the minimum on time limit.
                SUBWF   SW_C_CNT,W              ; Has the button been pressed a while ?
                GOTO    RETURN_FLAG

CHK_RIT_BRIEF:  MOVF    SW_BC_OLD,W
                MOVWF   TEMP_1                  ; Save the old button state.
                SUBWF   SW_BC_CNT,W             ; Has the button been released ?
                MOVF    SW_BC_CNT,W             ;    (current-old)
                MOVWF   SW_BC_OLD               ;    (copy current state to old state )
                GOTO    CHECK_TIMING

CHK_CAL_BRIEF:  MOVF    SW_B_OLD,W
                MOVWF   TEMP_1                  ; Save the old button state.
                SUBWF   SW_B_CNT,W              ; Has the button been released ?
                MOVF    SW_B_CNT,W              ;    (current-old)
                MOVWF   SW_B_OLD                ;    (copy current state to old state )
CHECK_TIMING:   BTFSC   CARRY
                GOTO    FALSE_FLAG              ; N. Return a false flag.
                MOVF    TEMP_1,W                ; Y. Was the button on less than min ?
                SUBLW   MIN_ON                  ;    (min on period-old)
                BTFSC   CARRY                   ;
                GOTO    FALSE_FLAG              ;    Y. Return a false flag.
                MOVF    TEMP_1,W                ;    N. Was the button on less than max ?
                SUBLW   MAX_ON                  ;          (max on period-old)
RETURN_FLAG:    BSF     PRESSED                 ;       Y. Return true flag.
                BTFSS   CARRY
FALSE_FLAG:     BCF     PRESSED                 ;       N. Return false flag.
                RETURN

;*****************************************************************************************
;
;      NAME:    MULTIPLY
;
;   PURPOSE:    40 bit * 40 bit unsigned multiply.    (ARG1 = ARG2 x ARG3)
;
;     INPUT:    ARG3_4..0  (MSD..LSD)
;               ARG2_4..0  (MSD..LSD)
;
;    OUTPUT:    ARG1_9..0  (MSD..LSD)
;               (ARG2 and ARG3 unchanged)
;
;*****************************************************************************************

MULTIPLY:       BTFSS   ARG2_4,7                ; Is ARG2 negative ?
                GOTO    MULT                    ; N. Do the multiply.
                CALL    NEG_ARG2                ; Y. Convert to positive.
                CALL    MULT                    ;    Do the multiply.
                GOTO    NEG_ARG2                ;    Convert back to negative.
                                                ;    Return via "GOTO".

MULT:           CALL    CLEAR_ARG1
                MOVLW   40                      ; 40 bits to multiply.
                MOVWF   COUNT_1                 ;
MLOOP:          RRF     ARG3_0,W                ; Rotate the multiplier right 1 bit.
                RRF     ARG3_4,F                ;
                RRF     ARG3_3,F                ;
                RRF     ARG3_2,F                ; ( after 40 rotates ARG3 will be )
                RRF     ARG3_1,F                ; ( returned to its original value )
                RRF     ARG3_0,F                ;

                BTFSC   CARRY                   ; Need to add?
                CALL    ADD_40                  ; Y. ARG1_9..5 = ARG1_9..5 + ARG2_4..0.

                RRF     ARG1_9,F                ; Right shift the partial product.
                RRF     ARG1_8,F                ;
                RRF     ARG1_7,F                ;
                RRF     ARG1_6,F                ;
                RRF     ARG1_5,F                ;
                RRF     ARG1_4,F                ;
                RRF     ARG1_3,F                ;
                RRF     ARG1_2,F                ;
                RRF     ARG1_1,F                ;
                RRF     ARG1_0,F                ;

                DECFSZ  COUNT_1,F               ; All 40 bits done ?
                GOTO    MLOOP                   ; N. Keep going.
                RETURN                          ; Y. All done.

ADD_40          MOVF    ARG2_0,W                ; 40 bit addition
                ADDWF   ARG1_5, F
                MOVF    ARG2_1,W                ; ARG1_9..5 = ARG1_9..5 + ARG2_4..0
                BTFSC   CARRY
                INCFSZ  ARG2_1,W
                ADDWF   ARG1_6, F
                MOVF    ARG2_2,W
                BTFSC   CARRY
                INCFSZ  ARG2_2,W
                ADDWF   ARG1_7, F
                MOVF    ARG2_3,W
                BTFSC   CARRY
                INCFSZ  ARG2_3,W
                ADDWF   ARG1_8, F
                MOVF    ARG2_4,W
                BTFSC   CARRY
                INCFSZ  ARG2_4,W
                ADDWF   ARG1_9, F
                RETURN

;*****************************************************************************************
;
;      NAME:    CHK_DDS
;
;   PURPOSE:    Update the AD9850 DDS chip if the frequency has changed.
;
;               Sends a 40 bit control word to the AD9850 DDS chip in serial form.
;               32 bit frequency + 8 bit phase/power down.
;               LSB of frequency is sent first.
;
;     INPUT:    ARG2 = Frequency.
;               ARG3 = DDS cal constant.  (2^56 / XTAL FREQ)
;
;    OUTPUT:    None.
;               (ARG2 and ARG3 unchanged)
;
;*****************************************************************************************

CHK_DDS:        CALL    PUSH_ARG3               ; Dont destroy ARG3.
                MOVLW   DDS_OLD                 ; Get the previous DDS freq.
                CALL    COPY_TO_ARG3

                CALL    CMP                     ; Is it the same as the current freq ?
                BTFSC   ZERO
                GOTO    PULL_ARG3               ; Y. Restore ARG3 then return via "GOTO".
                                                ;    (dot update the DDS chip)

                CALL    PULL_ARG3               ; N. Recover the original value of ARG3.

                MOVLW   DDS_OLD                 ;    Save the new DDS freq.
                CALL    COPY_FROM_ARG2

                CALL    MULTIPLY                ;    Mult DDS freq by constant.
                                                ;    (result ARG1_6..3)
SEND_DDS_WORD:  CLRF    ARG1_7                  ; Clear the phase/power down byte.
                MOVLW   40                      ; 40 bits to send.
                MOVWF   COUNT_1
NEXT_BIT:       RRF     ARG1_7,F                ; Shift right the 40 bits,
                RRF     ARG1_6,F                ; to move the LSB into the carry.
                RRF     ARG1_5,F
                RRF     ARG1_4,F
                RRF     ARG1_3,F
                BTFSS   CARRY                   ; Is the carry bit 1 ?
                BCF     DDS_DATA                ; N. Send a 0.
                BTFSC   CARRY                   ; Is the carry bit 1 ?
                BSF     DDS_DATA                ; Y. Send a 1.
                BSF     DDS_CLOCK               ; Toggle the DDS word load clock pin.
                BCF     DDS_CLOCK
                DECFSZ  COUNT_1,F               ; Have all the bits been sent?
                GOTO    NEXT_BIT                ; N. Keep going.
                BSF     DDS_LOAD                ; Toggle the DDS frequency update pin.
                BCF     DDS_LOAD
                RETURN

;*****************************************************************************************
;
;      NAME:    BIN_TO_DEC
;
;   PURPOSE:    Convert a 40 bit binary number to a 11 digit ASCII decimal number.
;               First the 40 bit binary number is converted to 11 digit decimal.
;               It is then converted to ASCII, including negative sign, decimal point
;               and leading zero blanking, ready for display on the LCD.
;               (if used on 10 MHz CAL screen, do not include the decimal point.)
;
;     INPUT:    ARG2 = frequency.
;
;    OUTPUT:    11 digit ASCII decimal number in ARG1_10..0. (MSD..LSD)
;               (ARG2 unchanged)
;
;*****************************************************************************************

BIN_TO_DEC:     CALL    CLEAR_ARG1
                MOVLW   40
                MOVWF   COUNT_2                 ; 40 bits to process.
                GOTO    SHIFT_TO_DEC

LOOP40:         CALL    SET_ARG1                ; Set FSR = ARG1_10, COUNT_1 = 11.
DEC_LOOP:       MOVLW   3
                ADDWF   INDF,W                  ; If num + 3 > 7 then num = num + 3.
                MOVWF   TEMP_1                  ; Add 3 to num. (adds 6, after next shift)
                BTFSS   TEMP_1,3                ; Is result > 7 ? (bit 3 set)
                GOTO    NO_ADJ
                ADDLW   B'01111000'             ; Y. Move bit 3 to bit 7.
                MOVWF   INDF                    ;    Put the number back in the buffer.
NO_ADJ:         INCF    FSR,F                   ; Point to next byte in buffer.
                DECFSZ  COUNT_1,F
                GOTO    DEC_LOOP                ; Loop until all 11 bytes done.

SHIFT_TO_DEC:   RLF     ARG2_4,W                ; Shift 40 bit number left by one bit
                RLF     ARG2_0,F                ; into dec buffer.
                RLF     ARG2_1,F
                RLF     ARG2_2,F                ; After 40 shifts ARG2 unchanged.
                RLF     ARG2_3,F
                RLF     ARG2_4,F                ; MSD of ARG2
                RLF     ARG1_0,F                ; LSD of ARG1
                RLF     ARG1_1,F
                RLF     ARG1_2,F
                RLF     ARG1_3,F
                RLF     ARG1_4,F
                RLF     ARG1_5,F
                RLF     ARG1_6,F
                RLF     ARG1_7,F
                RLF     ARG1_8,F
                RLF     ARG1_9,F
                RLF     ARG1_10,F               ; MSD of ARG1.
                DECFSZ  COUNT_2,F
                GOTO    LOOP40                  ; Loop until all 40 bits processed.

                BCF     LEADING_FLAG            ; Blank leading zeros until flag is set.

                CALL    SET_ARG1                ; Set FSR = ARG1_10, COUNT_1 = 11.
ASCII_LOOP:     MOVF    INDF,W                  ; Get the number.
                BTFSC   ZERO                    ; Is it = 0 ?
                GOTO    BLANK
NO_BLANK        ADDLW   H'30'                   ; N. Convert it to ASCII.
                MOVWF   INDF
                BSF     LEADING_FLAG            ;    Set the leading blank flag,
                GOTO    NUM_DONE                ;    following zeros will not be blanked.
BLANK:          BTFSC   LEADING_FLAG            ; Y. Is the leading blank flag set.
                GOTO    NO_BLANK                ;    Y. Convert it to ASCII.
                MOVLW   ' '
                MOVWF   INDF                    ;    N. Blank the zero.
                MOVF    FSR,W                   ;       Save the location of the blank
                MOVWF   COUNT_2                 ;       so that a - sign can be inserted.
NUM_DONE:       INCF    FSR,F                   ; Point to the next byte.
                MOVLW   8
                SUBWF   COUNT_1,W               ; Time to cancel leading blanking?
                BTFSC   ZERO
                BSF     LEADING_FLAG            ; Y. Always display char before DP.
                DECFSZ  COUNT_1,F
                GOTO    ASCII_LOOP

                BTFSS   NEGATIVE                ; Was this a negative number?
                RETURN                          ; N. Just return.
                MOVF    COUNT_2,W               ; Y. Get position of last blank.
                MOVWF   FSR
                MOVLW   '-'                     ;    Place - sign in front of the number.
                MOVWF   INDF
                RETURN

;*****************************************************************************************
;
;      NAME:    FREQ_TO_LCD
;
;   PURPOSE:    Update the DDS chip if required, then add the offset freq and multiply
;               by DISP_MULT and display it on the LCD.
;
;     INPUT:    ARG2 = frequency.
;
;    OUTPUT:    None.
;               (ARG1 & ARG3 destroyed)
;
;*****************************************************************************************

FREQ_TO_LCD:    MOVLW   EE_DDS_CAL
                CALL    EE_TO_ARG3              ; Get the XTAL cal constant. (2^56/XTAL)
                CALL    CHK_DDS                 ; Update the DDS chip if required.

                MOVLW   EE_OFFSET_FREQ
                CALL    EE_TO_ARG3              ; Get the offset freq.
                CALL    ADD                     ; Add the offset.

                CALL    CLEAR_ARG3              ; Move the mult byte from EEPROM to ARG3.
                CALL    READ_MULT
                MOVWF   ARG3_0

                CALL    MULTIPLY                ; Do the multiply.
                MOVLW   ARG1_4                  ; Get the result.
                CALL    COPY_TO_ARG2            ; Display the frequency, by falling
                                                ; through to DISPLAY_FREQ.
                                                ; Return via DISPLAY_FREQ.

;*****************************************************************************************
;
;      NAME:    DISPLAY_FREQ
;
;   PURPOSE:    Display the frequency in ARG2 on the LCD.
;               Including negative sign, decimal point and leading zero blanking.
;               (if used on 10 MHz CAL screen, do not include the decimal point.)
;
;     INPUT:    ARG2 = frequency.
;
;    OUTPUT:    None.
;
;*****************************************************************************************

DISPLAY_FREQ:   BCF     NEGATIVE                ; Clear the negative flag.
                BTFSS   ARG2_4,7                ; Is ARG2 a negative number?
                GOTO    DISP_FREQ               ; N. Just continue.
                BSF     NEGATIVE                ; Y. Set the negative flag.
                CALL    NEG_ARG2                ;    Convert it to a + number.
DISP_FREQ:      CALL    BIN_TO_DEC              ; Convert the number to ASCII decimal.
                MOVLW   H'C0'                   ; Position the cursor, line 2 pos 1.
                BTFSC   CAL_SCRN                ; Are we in the CAL 10MHz screen ?
                MOVLW   H'C2'                   ; Y. Shift the display to line 2 pos 3.
                CALL    LCD_CMD
                CALL    SET_ARG1                ; Set FSR = ARG1_10, COUNT_1 = 11.
DISP_LOOP:      MOVF    INDF,W                  ; Get char pointed to by FSR.
                CALL    LCD_CHR                 ; Display the char.
                BTFSC   CAL_SCRN                ; Are we in the CAL 10MHz screen ?
                GOTO    NO_DP
                MOVLW   7
                SUBWF   COUNT_1,W               ; N. Is it time to insert a decimal point?
                BTFSC   ZERO
                CALL    LCD_DP                  ;    Y. Print a decimal point.
NO_DP:          INCF    FSR,F                   ; Inc the pointer.
                DECFSZ  COUNT_1,F               ; All done ?
                GOTO    DISP_LOOP               ; N. Continue.

                BTFSC   NEGATIVE                ; Was the number negative ?
                CALL    NEG_ARG2                ; Y. Return it to a negative number.

                BTFSC   CAL_SCRN                ; Are we in the CAL 10MHz screen ?
                RETURN                          ; Y. All done.
                MOVLW   MHZ
                GOTO    LCD_TEXT                ; N. Display ' MHz'.
                                                ;    Return via "GOTO".

;*****************************************************************************************
;
;      NAME:    INIT_LCD
;
;   PURPOSE:    Initialize the 16 x 2 liquid crystal display.
;               The LCD controller chip must be equivalent to the HITACHI 44780.
;
;     INPUT:    None.
;
;    OUTPUT:    None.
;
;*****************************************************************************************

INIT_LCD:       CALL    LCD_RESET               ; Reset the LCD.
                CALL    LCD_RESET               ; Reset the LCD.
                CALL    LCD_RESET               ; Reset the LCD.

                MOVLW   B'00101000'             ; LCD 4 bit mode command.
                MOVWF   PORTB                   ; Also sets LCD_E lead high.
                BCF     LCD_E
                CALL    WAIT_200US
                MOVLW   B'00101000'             ; Set LCD to 2 line, 5x7 dot.
                CALL    LCD_CMD
                MOVLW   B'00001000'             ; Turn the display off.
                CALL    LCD_CMD
                CALL    LCD_CLEAR               ; Clear the display.
                MOVLW   B'00000110'             ; Cursor increments, no display shift.
                CALL    LCD_CMD
LCD_NORM:       MOVLW   B'00001100'             ; Turn the display on.
                GOTO    LCD_CMD                 ; Return via "GOTO".

LCD_RESET:      MOVLW   B'00111000'             ; LCD reset command.
                MOVWF   PORTB                   ; Also sets LCD_E lead high.
                BCF     LCD_E
                GOTO    WAIT_8MS                ; Return via "GOTO".

;*****************************************************************************************
;
;     NAMES:    LCD_CLEAR       Clear the display and home the cursor.
;               LCD_HOME        Home the cursor.
;               LCD_DP          Display a decimal point.
;               LCD_CMD         Send the command byte in W to the LCD.
;               LCD_CHR         Send the data byte in W to the LCD.
;               WAIT_8MS        Do nothing for 8 mS.
;               MS_WAIT         W = The number of mS to wait.
;
;*****************************************************************************************

LCD_CLEAR:      MOVLW   H'01'                   ; Clear the display. Then wait 8mS.
                CALL    LCD_CMD
WAIT_8MS:       MOVLW   8                       ; Set the outer loop to 8mS.
MS_WAIT:        MOVWF   COUNT_0                 ; Save the number of mS delay.
OUTER_LOOP:     CLRW                            ; Set the wait loop to 256.
                CALL    WAIT_LOOP               ; 256 loops x 4 x 1uS = 1.024mS.
                DECFSZ  COUNT_0,F               ; Dec the mS counter.
                GOTO    OUTER_LOOP
                RETURN

;-----------------------------------------------------------------------------------------

LCD_HOME:       MOVLW   H'80'                   ; Position the cursor, line 1 pos 1.
LCD_CMD:        BCF     LCD_RS                  ; Select the instruction register.
                GOTO    LCD_BYTE                ; Send the byte to the LCD.
;-----------------------------------------------------------------------------------------

LCD_DP:         MOVLW   H'A5'                   ; Display a decimal point.
LCD_CHR:        BSF     LCD_RS                  ; Select display data register.
LCD_BYTE:       MOVWF   TEMP_3                  ; Save the byte to send to the LCD.
                CALL    SEND_NIBBLE             ; Send the high nibble to the LCD.
                SWAPF   TEMP_3,F                ; Send the low nibble to the LCD.
SEND_NIBBLE:    MOVLW   B'00001111'
                ANDWF   PORTB,F                 ; Clear the high nibble of port B.
                MOVF    TEMP_3,W                ; Get the byte of data.
                ANDLW   B'11110000'             ; Keep only the high nibble.
                IORWF   PORTB,F                 ; Send the nibble to port B.
                BSF     LCD_E                   ; Pulse the LCD_E lead.
                BCF     LCD_E
WAIT_200US:     MOVLW   256-50                  ; Set the wait loop to 200uS.
WAIT_LOOP:      ADDLW   1                       ; 1 CLK cycles \
                BTFSS   ZERO                    ; 1     "       > 4 CLKs X 1uS
                GOTO    WAIT_LOOP               ; 2     "      /
                RETURN

;*****************************************************************************************
;
;     NAMES:    DISP_RIT        Display 'RIT' on line 1 pos 14. Then display RX or TX.
;               DISP_TX_RX      Display 'RX ' or 'TX ' on line 1 pos 1.
;
;     INPUT:    None.
;
;    OUTPUT:    None.
;
;-----------------------------------------------------------------------------------------
;
;      NAME:    LCD_TEXT
;
;   PURPOSE:    Display on the LCD, the string of characters pointed to by W. 
;               The string must end with a 0 byte.
;
;     INPUT:    W = String offset address. (offset from the base of the text table)
;
;    OUTPUT:    None.
;
;*****************************************************************************************

DISP_RIT:       MOVLW   H'8D'
                CALL    LCD_CMD                 ; Position the cursor, line 1 pos 14.
                MOVLW   RIT_OFF
                CALL    LCD_TEXT                ; Display 'RIT'.
                                                ; Fall through and display RX or TX.
DISP_TX_RX:     CALL    LCD_HOME                ; Goto the start of line 1.
                CALL    CHK_TX                  ; Check the TX button.
                MOVLW   RX                      ; Prepare to display 'RX '.
                BTFSC   PRESSED                 ; Was the TX button pressed ?
                MOVLW   TX                      ; Y. Prepare to display 'TX '.
                                                ; Display the 'TX ' or 'RX '.

LCD_TEXT:       MOVWF   TEMP_1                  ; Save the text start address.
                CALL    TABLE
                ANDLW   H'FF'
                BTFSC   ZERO                    ; At end of message ? (0 returned at end)
                RETURN                          ; Y. All done.
                CALL    LCD_CHR                 ; N. Display character.
                INCF    TEMP_1,W                ;    Point to next character.
                GOTO    LCD_TEXT

;*****************************************************************************************


                END
